;ordinal arithmetic definitions 
(in-package "ACL2")

(defun natp (x)
  (declare (xargs :guard t))
  "Recognizer for natural numbers"
  (and (integerp x)
       (<= 0 x)))

(defun posp (x)
  "Recognizer for positive integers"
  (declare (xargs :guard t))
  (and (integerp x)
       (< 0 x)))

(defun first-exp (x) 
  "The exponent of the first term of a cnf ordinal"
  (declare (xargs :guard (or (atom x) (consp (car x)))))
  (if (atom x) 
      0 
    (caar x)))

(defun first-coef (x)
  "The coefficient of the first term of a cnf ordinal"
  (declare (xargs :guard (or (atom x) (consp (car x)))))
  (if (atom x) 
      x 
    (cdar x)))

(defmacro first-term (x)
  "gives the first term of x"
  `(cons (cons (first-exp ,x) (first-coef ,x)) 0))

(defun omega-term (x k)
  "Creates (w^x)*k"
  (declare (xargs :guard t))
  (cons (cons x k) 0))

(defun o<g (x)
  "Used *only* for the guard proofs"
  (declare (xargs :guard t))
  (if (atom x)
      (rationalp x)
    (and (consp (car x))
         (rationalp (cdar x))
         (o<g (caar x))
         (o<g (cdr x)))))

(defun o< (x y)
  "The less-than relation on cnf ordinals"
  (declare (xargs :guard (and (o<g x) (o<g y))))
  (cond ((atom x) 
         (or (consp y) (< x y)))
        ((atom y) nil)
        ((not (equal (first-exp x) (first-exp y))) 
         (o< (first-exp x) (first-exp y)))
        ((not (= (first-coef x) (first-coef y))) 
         (< (first-coef x) (first-coef y)))
        (t (o< (cdr x) (cdr y)))))

(defun op (a)
  "A recognizer for cnf ordinals"
  (declare (xargs :guard (o<g a))) 
  (if (atom a)
      (natp a)
    (and (consp (car a))
         (op (first-exp a))
         (not (eql 0 (first-exp a)))
         (posp (first-coef a))
         (op (cdr a))
	 (o< (first-exp (cdr a))
	     (first-exp a)))))

(defun opg (a)
  "Used *only* for the guard proofs"
  (declare (xargs :guard t))
  (if (atom a)
      (natp a)
    (and (o<g a)
         (op a))))

(defthm op-implies-o<g
  (implies (op a)
           (o<g a)))

(defthm opg-is-op
  (equal (opg a)
         (op a)))

(in-theory (disable opg))

(defmacro omega ()
  '(quote ((1 . 1) . 0)))

(defun o-max (a b)
  (declare (xargs :guard (and (opg a) (opg b))))
  (if (o< a b) b a))

(defun limit-cdr-p (a)
  (declare (xargs :guard (opg a)))
  (if (consp a)
      (limit-cdr-p (cdr a))
    (= a 0)))

(defun limitp (a)
  (declare (xargs :guard (opg a)))
  (and (consp a)
       (op a)
       (limit-cdr-p (cdr a))))

(defun ob+ (x y)
  "Ordinal addition"
  (declare (xargs :guard (and (opg x) (opg y))))
  (let ((x0 (first-exp x)) (y0 (first-exp y))
	(x1 (first-coef x)) (y1 (first-coef y)))
    (cond ((and (atom x) (atom y)) (+ x y))
	  ((or (atom x) (o< x0 y0)) y)
	  ((o< y0 x0) (cons (cons x0 x1) 
			    (ob+ (cdr x) y)))
	  (t (cons (cons x0 (+ x1 y1))
		   (cdr y))))))

(defmacro o+ (&rest rst)
; based on the macro +
  (if rst
      (if (cdr rst)
          (xxxjoin 'ob+ rst)
        (car rst))
    0))

(defun o- (x y)
  "Ordinal subtraction"
  (declare (xargs :guard (and (opg x) (opg y))))
  (let ((x0 (first-exp x)) (y0 (first-exp y))
	(x1 (first-coef x)) (y1 (first-coef y)))
    (cond ((o< x y) 0)
	  ((and (atom x) (atom y)) (- x y))
	  ((or (atom y) (o< y0 x0)) x)
	  ((< y1 x1) (cons (cons x0 (- x1 y1))
			   (cdr x)))
	  (t (o- (cdr x) (cdr y))))))

(defun ob* (x y)
  "Ordinal multiplication"
  (declare (xargs :guard (and (opg x) (opg y))))
  (let ((x0 (first-exp x)) (y0 (first-exp y))
	(x1 (first-coef x)) (y1 (first-coef y)))
    (cond ((or (equal x 0) (equal y 0)) 0)
	  ((and (atom x) (atom y)) (* x y))
	  ((atom y) 
	   (cons (cons x0 (* x1 y1)) 
		 (cdr x)))
	  (t (cons (cons (o+ x0 y0)
			 y1) 
		   (ob* x (cdr y)))))))

(defmacro o* (&rest rst)
; based on the macro *
  (cond ((null rst) 1)
        ((null (cdr rst))
         (car rst))
        (t (xxxjoin 'ob* rst))))

(defun ob^ (x y)
  "Ordinal exponentiation"
  (declare (xargs :guard (and (opg x) (opg y))
                  :verify-guards nil))
  (let ((x0 (first-exp x)) 
	(y0 (first-exp y))
	(y1 (first-coef y)))
    (cond ((or (and (atom y) 
		    (not (posp y))) ;(zp y))
	       (equal x 1))
	   1)
	  ((equal x 0)
	   0)
	  ((atom y)
	   (o* x (ob^ x (1- y))))
	  ((atom x)
	   (if (equal y0 1)
	       (o* (omega-term y1 1)
		   (ob^ x (cdr y)))
	     (o* (omega-term (omega-term (o- y0 1) 
					 y1)
			     1)
		 (ob^ x (cdr y)))))
	  (t 
	   (o* (omega-term (o* x0 (first-term y))
			   1)
	       (ob^ x (cdr y)))))))

(defmacro o^ (&rest rst)
; based on the macro *
  (cond ((null rst) 1)
        ((null (cdr rst))
         (car rst))
        (t (xxxjoin 'ob^ rst))))

(add-binop o+ ob+)
(add-binop o* ob*)
(add-binop o^ ob^)

